use crate::ast::*;
use crate::tokeniser::Keyword;
use crate::tokeniser::Token;

fn parse_basic(tokens: &Vec<Token>, i: &mut usize) -> Expression {
    let literal = match &tokens[*i] {
        Token::Identifier(identifier) => Expression::Variable(identifier.to_owned()),
        Token::Number(number) => Expression::Literal(Literal::Int(number.parse::<i64>().unwrap())),
        _other => {
            dbg!(&i);
            dbg!(_other);
            todo!()
        }
    };
    *i += 1;
    literal
}

// TODO: replace parse_math with this recursive function
pub fn parse_expression(tokens: &Vec<Token>, i: &mut usize) -> Expression {
    #[derive(Debug)]
    enum OpExpr {
        Expression(Expression),
        BinOp(BinOp),
        Unary(Unary),
    }

    let start_i = *i;
    let mut in_expressions_and_operators: Vec<OpExpr> = Vec::new();

    loop {
        in_expressions_and_operators.push(match &tokens[*i] {
            Token::OpenParanthesis => {
                *i += 1;
                let expr = OpExpr::Expression(parse_expression(tokens, i));
                *i += 1;
                expr
            }
            Token::Plus => {
                *i += 1;
                OpExpr::BinOp(BinOp::Plus)
            }
            Token::Minus => {
                *i += 1;
                OpExpr::BinOp(BinOp::Minus)
            }
            Token::Ampersand => {
                *i += 1;
                OpExpr::Unary(Unary::ValueFrom)
            }
            Token::Star => {
                *i += 1;
                OpExpr::Unary(Unary::PointerTo)
            }
            Token::Number(_) => {
                dbg!(OpExpr::Expression(parse_basic(tokens, i)))
            }
            Token::Identifier(_) => OpExpr::Expression(parse_basic(tokens, i)),
            _other => break,
        });
    }

    dbg!(&i);
    dbg!(&in_expressions_and_operators);

    if in_expressions_and_operators.len() == 1 {
        return match &in_expressions_and_operators.pop().unwrap() {
            OpExpr::Expression(expression) => expression.to_owned(),
            _ => {
                todo!()
            }
        };
    }

    *i = start_i;

    let prioritizing_order_binops: Vec<Vec<BinOp>> = vec![vec![BinOp::Plus, BinOp::Minus]];

    let prioritizing_order_unaryops: Vec<Vec<Unary>> =
        vec![vec![Unary::ValueFrom, Unary::PointerTo]];

    let mut ordered_ops: Vec<(usize, bool)> = Vec::new();

    for prioritizing_order in prioritizing_order_binops {
        for j in (0..in_expressions_and_operators.len()).rev() {
            let operator = match &in_expressions_and_operators[j] {
                OpExpr::BinOp(binop) => binop,
                _ => continue,
            };

            if prioritizing_order.contains(operator) {
                ordered_ops.push((j, true));
            };
        }
    }

    for prioritizing_order in prioritizing_order_unaryops {
        for j in (0..in_expressions_and_operators.len()).rev() {
            let operator = match &in_expressions_and_operators[j] {
                OpExpr::Unary(unary) => unary,
                _other => continue,
            };

            if prioritizing_order.contains(operator) {
                ordered_ops.push((j, false));
            };
        }
    }

    let this_operator = ordered_ops.get(0);

    let this_operator = match this_operator {
        Some(val) => val.to_owned(),
        None => return parse_basic(tokens, i),
    };

    dbg!(this_operator);

    match this_operator {
        (index, true) => {
            let mut start_vec = Vec::new();

            for i in 0..(index + *i) {
                start_vec.push(tokens[i].clone());
            }

            start_vec.push(Token::Semi);
            dbg!(&start_vec);

            let start = parse_expression(&start_vec, i);
            // let start = Expression::Literal(Literal::Int(42));

            dbg!(&start);

            *i += 1;

            dbg!(&i);
            let end = parse_expression(tokens, i);

            Expression::Binary(
                match &in_expressions_and_operators[index] {
                    OpExpr::BinOp(binop) => binop.to_owned(),
                    _other => {
                        dbg!(start);
                        dbg!(end);
                        dbg!(_other);
                        todo!()
                    }
                },
                Box::new(start),
                Box::new(end),
            )
        }
        (index, false) => {
            *i += 1;

            let end = parse_expression(tokens, i);

            Expression::Unary(
                match &in_expressions_and_operators[index] {
                    OpExpr::Unary(unary) => unary.to_owned(),
                    _other => {
                        dbg!(end);
                        dbg!(_other);
                        todo!()
                    }
                },
                Box::new(end),
            )
        }
    }
}

fn parse_statements(tokens: &Vec<Token>, i: &mut usize) -> Vec<Statement> {
    dbg!(&i);
    let mut statements = Vec::new();
    while *i < tokens.len() {
        statements.push(match &tokens[*i] {
            Token::OpenSquiglyBracket => {
                *i += 1;
                let statements = parse_statements(tokens, i);
                Statement::Scope(statements)
            }
            Token::Keyword(Keyword::Let) => {
                *i += 1;
                let name = match tokens[*i].clone() {
                    Token::Identifier(identifier) => identifier,
                    _ => todo!(),
                };

                *i += 1;
                match tokens[*i].clone() {
                    Token::Equals => {}
                    _ => todo!(),
                };

                *i += 1;
                Statement::VariableDeclaration(name, parse_expression(tokens, i))
            }
            Token::Keyword(Keyword::Assign) => {
                *i += 1;
                let name = match tokens[*i].clone() {
                    Token::Identifier(identifier) => identifier,
                    _ => todo!(),
                };

                *i += 1;
                match tokens[*i].clone() {
                    Token::Equals => {}
                    _ => todo!(),
                };

                *i += 1;

                Statement::VariableAssignment(name, parse_expression(tokens, i))
            }
            Token::Identifier(identifier) => {
                *i += 1;

                let mut arguments: Vec<Expression> = Vec::new();

                while tokens[*i] != Token::CloseParanthesis {
                    *i += 1;
                    if tokens[*i] == Token::CloseParanthesis {
                        break;
                    }
                    dbg!(&i);
                    dbg!(tokens);
                    dbg!(&tokens[*i]);
                    arguments.push(parse_expression(tokens, i));
                }

                *i += 1;

                dbg!(&arguments);

                Statement::FunctionCall(identifier.to_owned(), arguments)
            }
            Token::CloseSquiglyBracket => {
                *i += 1;
                return statements;
            }
            Token::Semi => {
                *i += 1;
                continue;
            }
            _other => {
                dbg!(_other);
                dbg!(i);
                todo!()
            }
        });
    }
    *i += 1;
    statements
}

pub fn parse_arguments(tokens: &Vec<Token>, i: &mut usize) -> Vec<String> {
    let mut argument_variables = Vec::new();
    while tokens[*i] != Token::OpenSquiglyBracket {
        match &tokens[*i] {
            Token::Identifier(identifier) => argument_variables.push(identifier.to_owned()),
            Token::OpenSquiglyBracket => {}
            _other => todo!(),
        }
        *i += 1;
    }
    argument_variables
}

pub fn ast(tokens: Vec<Token>) -> Vec<Function> {
    let mut functions: Vec<Function> = Vec::new();

    let mut i = 0;
    while i < tokens.len() {
        match &tokens[i] {
            Token::Keyword(Keyword::Function) => {
                i += 1;

                let name = match &tokens[i] {
                    Token::Identifier(identifier) => identifier,
                    _ => todo!(),
                }
                .to_owned();
                i += 1;

                let arguments = parse_arguments(&tokens, &mut i);

                i += 1;
                let statements = parse_statements(&tokens, &mut i);

                let statement = Statement::Scope(statements);

                functions.push(Function::new(name, arguments, statement));
            }
            _other => {
                dbg!(&functions);
                dbg!(i);
                dbg!(_other);
                todo!();
            }
        };

        i += 1;
    }

    functions
}
